@@ -4,20 +4,16 @@
  * For license terms, see the file COPYING in this directory.
  */
 #include "config.h"
+#include "util.h"
+#include "fetchmail.h"
 
 #include <stdio.h>
-#if defined(STDC_HEADERS)
 #include <stdlib.h>
-#endif
-#if defined(HAVE_UNISTD_H)
 #include <unistd.h>
-#endif
 #include <fcntl.h>
 #include <string.h>
 #include <signal.h>
-#if defined(HAVE_SYSLOG)
 #include <syslog.h>
-#endif
 #include <pwd.h>
 #ifdef __FreeBSD__
 #include <grp.h>
@@ -25,25 +21,20 @@
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#ifdef HAVE_SETRLIMIT
 #include <sys/resource.h>
-#endif /* HAVE_SETRLIMIT */
 
 #ifdef HAVE_SOCKS
 #include <socks.h> /* SOCKSinit() */
 #endif /* HAVE_SOCKS */
 
-#ifdef HAVE_LANGINFO_H
 #include <langinfo.h>
-#endif
 
-#include "fetchmail.h"
-#include "socket.h"
 #include "tunable.h"
 #include "smtp.h"
 #include "netrc.h"
 #include "i18n.h"
 #include "lock.h"
+#include "getopt.h" // for optind
 
 /* need these (and sys/types.h) for res_init() */
 #include <netinet/in.h>
@@ -89,10 +80,11 @@ static int querystatus;		/* status of query */
 static int successes;		/* count number of successful polls */
 static int activecount;		/* count number of active entries */
 static struct runctl cmd_run;	/* global options set from command line */
-static time_t parsetime;	/* time of last parse */
+static time_t parsetime;	/* time of last parse of fetchmail's rcfile */
+static time_t netrc_parsetime;  /* time of last parse of .netrc file */
 
-static RETSIGTYPE terminate_run(int);
-static RETSIGTYPE terminate_poll(int);
+static void terminate_run(int);
+static void terminate_poll(int);
 
 #if defined(__FreeBSD__) && defined(__FreeBSD_USE_KVM)
 /* drop SGID kmem privileage until we need it */
@@ -115,7 +107,7 @@ static void dropprivs(void)
 }
 #endif
 
-#if defined(HAVE_SETLOCALE) && defined(ENABLE_NLS) && defined(HAVE_STRFTIME)
+#if defined(ENABLE_NLS)
 #include <locale.h>
 /** returns timestamp in current locale,
  * and resets LC_TIME locale to POSIX. */
@@ -134,7 +126,7 @@ static char *timestamp (void)
 #define timestamp rfc822timestamp
 #endif
 
-static RETSIGTYPE donothing(int sig) 
+static void donothing(int sig) 
 {
     set_signal_handler(sig, donothing);
     lastsig = sig;
@@ -146,8 +138,8 @@ static void printcopyright(FILE *fp) {
 		   "                   Robert M. Funk, Graham Wilson\n"
 		   "Copyright (C) 2005 - 2012 Sunil Shetye\n"
 		   "Copyright (C) 2005 - %d Matthias Andree\n"
-		   ), 2022);
-	fprintf(fp, GT_("Fetchmail comes with ABSOLUTELY NO WARRANTY. This is free software, and you\n"
+		   ), 2025);
+	fprintf(fp, "%s", GT_("Fetchmail comes with ABSOLUTELY NO WARRANTY. This is free software, and you\n"
 		   "are welcome to redistribute it under certain conditions. For details,\n"
 		   "please see the file COPYING in the source or documentation directory.\n"));
 #if defined(SSL_ENABLE) && !defined(USING_WOLFSSL)
@@ -156,8 +148,69 @@ static void printcopyright(FILE *fp) {
 #endif
 }
 
+static void reportquerystatus(FILE *fp) {
+	const char *desc;
+	switch(querystatus)
+	{
+	case PS_SUCCESS:
+	    desc = "SUCCESS"; break;
+	case PS_NOMAIL:
+	    desc = "NOMAIL"; break;
+	case PS_SOCKET:
+	    desc = "SOCKET"; break;
+	case PS_AUTHFAIL:
+	    desc = "AUTHFAIL"; break;
+	case PS_PROTOCOL:
+	    desc = "PROTOCOL"; break;
+	case PS_SYNTAX:
+	    desc = "SYNTAX"; break;
+	case PS_IOERR:
+	    desc = "IOERR"; break;
+	case PS_ERROR:
+	    desc = "ERROR"; break;
+	case PS_EXCLUDE:
+	    desc = "EXCLUDE"; break;
+	case PS_LOCKBUSY:
+	    desc = "LOCKBUSY"; break;
+	case PS_SMTP:
+	    desc = "SMTP"; break;
+	case PS_DNS:
+	    desc = "DNS"; break;
+	case PS_BSMTP:
+	    desc = "BSMTP"; break;
+	case PS_MAXFETCH:
+	    desc = "MAXFETCH"; break;
+	default:
+	    desc = NULL;
+	}
+
+	if (desc) {
+	    report(fp, GT_("Query status=%d (%s)\n"), querystatus, desc);
+	} else {
+	    report(fp, GT_("Query status=%d\n"), querystatus);
+	}
+}
+
 const char *iana_charset;
 
+/**
+ * @brief check time_t sanity and log a warning 10 years before Y2038 becomes a problem.
+ *
+ * Just in case someone tries to run this in a computer museum and forgets stuffing
+ * things in a time machine.
+ */
+static void y2038_check_or_warn(void)
+{
+        const int epoch_ts_2028_01_01 = 1830297600 / 2;  /* 2028-01-01T00:00Z */
+        time_t t;
+        errno = 0;
+        if (time(&t) == (time_t)-1 && errno == EOVERFLOW) {
+                report(stderr, GT_("%s: WARNING: time() overflowed. Proper function is not guaranteed. Obtain an operating system which supports operation beyond the year 2038.\n"), program_name);
+        } else if (sizeof(time_t) < 8 && (time_t)-1 < 0 && time(NULL) >= epoch_ts_2028_01_01) { // four-byte time_t and time_t is signed} -> start warning 10 years before time
+                report(stderr, GT_("%s: WARNING: the operation system this program was built for does not function beyond the year 2038. Plan your update accordingly.\n"), program_name);
+        }
+}
+
 int main(int argc, char **argv)
 {
     int bkgd = FALSE;
@@ -165,10 +218,10 @@ int main(int argc, char **argv)
     flag safewithbg = FALSE; /** if parsed options are compatible with a
 			      fetchmail copy running in the background */
     struct query *ctl;
-    netrc_entry *netrc_list;
-    char *netrc_file, *tmpbuf;
+    char *netrc_file; /** long-lived variable containing the name of the .netrc path */
+    char *tmpbuf;
     pid_t pid;
-    int lastsig = 0;
+    lastsig = 0;
 
 #if defined(__FreeBSD__) && defined(__FreeBSD_USE_KVM)
     dropprivs();
@@ -176,6 +229,9 @@ int main(int argc, char **argv)
 
     envquery(argc, argv);
 #ifdef ENABLE_NLS
+#if defined(_CLANGD) && !defined(LOCALEDIR)
+#define LOCALEDIR "po/"
+#endif
     setlocale (LC_ALL, "");
     bindtextdomain(PACKAGE, LOCALEDIR);
     textdomain(PACKAGE);
@@ -268,22 +324,16 @@ int main(int argc, char **argv)
 	"-ODMR"
 #endif /* ODMR_ENABLE */
 #ifndef SSL_ENABLE
-	"-SSL"
+	"-TLS"
 #else
-#ifdef USING_WOLFSSL
+# ifdef USING_WOLFSSL
 	"+WOLFSSL"
-#else
-	"+SSL-SSLv2"
-#endif
-#if (HAVE_DECL_SSLV3_CLIENT_METHOD + 0 == 0) || defined(OPENSSL_NO_SSL3)
-	"-SSLv3"
-#endif
-#if HAVE_DECL_TLS1_2_VERSION + 0 == 0
-	"-TLS1.2"
-#endif
-#if HAVE_DECL_TLS1_3_VERSION + 0 == 0
+# else
+	"+TLS"
+# endif
+# if HAVE_DECL_TLS1_3_VERSION + 0 == 0
 	"-TLS1.3"
-#endif
+# endif
 #endif
 #ifdef OPIE_ENABLE
 	"+OPIE"
@@ -314,11 +364,15 @@ int main(int argc, char **argv)
 		   "Run-time uses SSL library %#lx \"%s\"\n"),
 			OPENSSL_VERSION_NUMBER, OPENSSL_VERSION_TEXT,
 			OpenSSL_version_num(), OpenSSL_version(OPENSSL_VERSION));
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L // OpenSSL_version API in this form added in OpenSSL 3.0 according to manpage
+# if defined(OPENSSL_DIR) && defined(OPENSSL_ENGINES_DIR) // in-depth fix in case some OpenSSL freerider claims 3.x compat without delivering it
 	printf(GT_("OpenSSL: %s\nEngines: %s\n"),
 			OpenSSL_version(OPENSSL_DIR),
 			OpenSSL_version(OPENSSL_ENGINES_DIR));
+# endif
+#endif
 # if !HAVE_DECL_TLS1_3_VERSION || defined(OPENSSL_NO_TLS1_3)
-	printf(GT_("WARNING: Your SSL/TLS library does not support TLS v1.3.\n"));
+# error Your SSL/TLS library does not support TLS v1.3.
 # endif
 # if defined(LIBRESSL_VERSION_NUMBER) && !defined(__OpenBSD__)
 	/* OpenBSD ships LibreSSL as part of the base system, so is exempt
@@ -332,13 +386,6 @@ int main(int argc, char **argv)
 	puts("");
 	printcopyright(stdout);
 	puts("");
-	fputs("Fallback MDA: ", stdout);
-#ifdef FALLBACK_MDA
-	fputs(FALLBACK_MDA, stdout);
-#else
-	fputs("(none)", stdout);
-#endif
-	putchar('\n');
 	fflush(stdout);
 
 	/* this is an attempt to help remote debugging */
@@ -379,16 +426,10 @@ int main(int argc, char **argv)
 	}
     }
 
-#if defined(HAVE_SYSLOG)
     /* logging should be set up early in case we were restarted from exec */
     if (run.use_syslog)
     {
-#if defined(LOG_MAIL)
 	openlog(program_name, LOG_PID, LOG_MAIL);
-#else
-	/* Assume BSD4.2 openlog with two arguments */
-	openlog(program_name, LOG_PID);
-#endif
 	/* precedence: logfile (if effective) overrides syslog. */
 	if (run.logfile) {
 	    syslog(LOG_ERR, GT_("syslog and logfile options are both set, ignoring syslog, and logging to %s"), run.logfile);
@@ -399,9 +440,10 @@ int main(int argc, char **argv)
 	}
     }
     else
-#endif
 	report_init((run.poll_interval == 0 || nodetach) && !run.logfile); /* when changing this, change copy above, too */
 
+    y2038_check_or_warn(); // this goes between initialization of logging and before we're getting serious
+
 #ifdef POP3_ENABLE
     /* initialize UID handling */
     {
@@ -420,7 +462,6 @@ int main(int argc, char **argv)
     /* construct the lockfile */
     fm_lock_setup(&run);
 
-#ifdef HAVE_SETRLIMIT
     /*
      * Before getting passwords, disable core dumps unless -v -d0 mode is on.
      * Core dumps could otherwise contain passwords to be scavenged by a
@@ -433,14 +474,20 @@ int main(int argc, char **argv)
 	corelimit.rlim_max = 0;
 	setrlimit(RLIMIT_CORE, &corelimit);
     }
-#endif /* HAVE_SETRLIMIT */
 
-#define	NETRC_FILE	".netrc"
     /* parse the ~/.netrc file (if present) for future password lookups. */
     netrc_file = prependdir (NETRC_FILE, home);
-    netrc_list = parse_netrc(netrc_file);
-    free(netrc_file);
-#undef NETRC_FILE
+    netrc_entry *netrc_list;
+    netrc_list = netrc_parse(netrc_file);
+    netrc_parsetime = time(NULL);
+    { // do not put this under netrc_list - if .netrc were empty, netrc_list would be NULL.
+        // note the parse time of netrc
+        struct stat st;
+        if (0 == stat(netrc_file, &st)) {
+                netrc_parsetime = st.st_mtime;
+        }
+    }
+    /* do NOT free(netrc_file) -- we need it later in the main loop for checking if its st_mtime changed */
 
     /* pick up passwords where we can */ 
     for (ctl = querylist; ctl; ctl = ctl->next)
@@ -453,29 +500,51 @@ int main(int argc, char **argv)
 		ctl->password = ctl->remotename;
 	    else
 	    {
-		netrc_entry *p;
+		const netrc_entry *p;
 
 		/* look up the pollname and account in the .netrc file. */
-		p = search_netrc(netrc_list,
+		p = netrc_search(netrc_list,
 				 ctl->server.pollname, ctl->remotename);
 		/* if we find a matching entry with a password, use it */
 		if (p && p->password)
+		{
 		    ctl->password = xstrdup(p->password);
-
+		}
 		/* otherwise try with "via" name if there is one */
 		else if (ctl->server.via)
 		{
-		    p = search_netrc(netrc_list, 
+		    p = netrc_search(netrc_list,
 				     ctl->server.via, ctl->remotename);
 		    if (p && p->password)
 		        ctl->password = xstrdup(p->password);
 		}
 	    }
 	}
+        if (ctl->active && !(implicitmode && ctl->server.skip) && ctl->server.esmtp_name && !ctl->server.esmtp_password)
+        {
+                // get ESMTP passwords from .netrc - the user names need to be in .fetchmailrc
+                for (struct idlist *i = ctl->smtphunt; i; i = i->next) {
+                        char *instring = xstrdup(i->id);
+			(void)strtok(instring, "/");  // chop off port from hostname
+                        const netrc_entry *entry = netrc_search(netrc_list, instring, ctl->server.esmtp_name);
+                        if (entry && entry->password) {
+                                ctl->server.esmtp_password = xstrdup(entry->password);
+                                xfree(instring);
+                                break;
+                        }
+                        xfree(instring);
+                }
+        }
+	if (ctl->moveto != NULL && ctl->server.protocol != P_IMAP)
+	{
+	    (void) fprintf(stderr,
+			   GT_("fetchmail: configuration invalid, --moveto is only valid for IMAP servers\n"));
+	    exit(PS_SYNTAX);
+	}
     }
 
-    free_netrc(netrc_list);
-    netrc_list = 0;
+    netrc_free(netrc_list);
+    netrc_list = NULL;
 
     /* perhaps we just want to check options? */
     if (versioninfo)
@@ -556,59 +625,63 @@ int main(int argc, char **argv)
     /* another fetchmail is running -- wake it up or die */
     if (pid != 0)
     {
-	if (check_only)
-	{
-	    fprintf(stderr,
-		 GT_("fetchmail: can't check mail while another fetchmail to same host is running.\n"));
-	    return(PS_EXCLUDE);
-        }
-	else if (!implicitmode)
-	{
-	    fprintf(stderr,
-		 GT_("fetchmail: can't poll specified hosts with another fetchmail running at %ld.\n"),
-		 (long)pid);
-		return(PS_EXCLUDE);
-	}
-	else if (!bkgd)
-	{
-	    fprintf(stderr,
-		 GT_("fetchmail: another foreground fetchmail is running at %ld.\n"),
-		 (long)pid);
-		return(PS_EXCLUDE);
-	}
-	else if (getpid() == pid)
+	if (getpid() == pid) {
 	    /* this test enables re-execing on a changed rcfile */
 	    fm_lock_assert();
-	else if (argc > 1 && !safewithbg)
-	{
-	    fprintf(stderr,
-		    GT_("fetchmail: can't accept options while a background fetchmail is running.\n"));
+	} else {
+	    xfree(netrc_file);
+	    if (check_only)
 	    {
-		int i;
-		fprintf(stderr, "argc = %d, arg list:\n", argc);
-		for (i = 1; i < argc; i++) fprintf(stderr, "arg %d = \"%s\"\n", i, argv[i]);
-	    }
-	    return(PS_EXCLUDE);
-	}
-	else if (kill(pid, SIGUSR1) == 0)
-	{
-	    if (outlevel > O_SILENT)
 		fprintf(stderr,
-			GT_("fetchmail: background fetchmail at %ld awakened.\n"),
+			GT_("fetchmail: can't check mail while another fetchmail to same host is running.\n"));
+		return (PS_EXCLUDE);
+	    }
+	    else if (!implicitmode)
+	    {
+		fprintf(stderr,
+			GT_("fetchmail: can't poll specified hosts with another fetchmail running at %ld.\n"),
 			(long)pid);
-	    return(0);
-	}
-	else
-	{
-	    /*
-	     * Should never happen -- possible only if a background fetchmail
-	     * croaks after the first kill probe above but before the
-	     * SIGUSR1/SIGHUP transmission.
-	     */
-	    fprintf(stderr,
-		    GT_("fetchmail: elder sibling at %ld died mysteriously.\n"),
-		    (long)pid);
-	    return(PS_UNDEFINED);
+		return (PS_EXCLUDE);
+	    }
+	    else if (!bkgd)
+	    {
+		fprintf(stderr,
+			GT_("fetchmail: another foreground fetchmail is running at %ld.\n"),
+			(long)pid);
+		return (PS_EXCLUDE);
+	    }
+	    else if (argc > 1 && !safewithbg)
+	    {
+		fprintf(stderr,
+			GT_("fetchmail: can't accept options while a background fetchmail is running.\n"));
+		{
+		    int i;
+		    fprintf(stderr, "argc = %d, arg list:\n", argc);
+		    for (i = 1; i < argc; i++)
+			fprintf(stderr, "arg %d = \"%s\"\n", i, argv[i]);
+		}
+		return (PS_EXCLUDE);
+	    }
+	    else if (kill(pid, SIGUSR1) == 0)
+	    {
+		if (outlevel > O_SILENT)
+		    fprintf(stderr,
+			    GT_("fetchmail: background fetchmail at %ld awakened.\n"),
+			    (long)pid);
+		return (0);
+	    }
+	    else
+	    {
+		/*
+		 * Should never happen -- possible only if a background fetchmail
+		 * croaks after the first kill probe above but before the
+		 * SIGUSR1/SIGHUP transmission.
+		 */
+		fprintf(stderr,
+			GT_("fetchmail: elder sibling at %ld died mysteriously.\n"),
+			(long)pid);
+		return (PS_UNDEFINED);
+	    }
 	}
     }
 
@@ -623,6 +696,7 @@ int main(int argc, char **argv)
 		fprintf(stderr,
 			GT_("fetchmail: can't find a password for %s@%s.\n"),
 			ctl->remotename, ctl->server.pollname);
+		xfree(netrc_file);
 		return(PS_AUTHFAIL);
 	    } else {
 		const char* password_prompt = GT_("Enter password for %s@%s: ");
@@ -632,6 +706,7 @@ int main(int argc, char **argv)
 		snprintf(tmpbuf, pplen, password_prompt,
 			ctl->remotename, ctl->server.pollname);
 		ctl->password = xstrdup((char *)fm_getpassword(tmpbuf));
+		fm_clearpassword();
 		free(tmpbuf);
 	    }
 	}
@@ -718,12 +793,15 @@ int main(int argc, char **argv)
     do {
 	/* 
 	 * Check to see if the rcfile has been touched.  If so,
-	 * re-exec so the file will be reread.  Doing it this way
+	 * re-exec so the file will be reread.	Doing it this way
 	 * avoids all the complications of trying to deallocate the
 	 * in-core control structures -- and the potential memory
 	 * leaks...
 	 */
 	struct stat	rcstat;
+	struct stat	netrcstat;
+	memset(&rcstat, 0, sizeof rcstat);
+	memset(&netrcstat, 0, sizeof netrcstat);
 
 	if (strcmp(rcfile, "-") == 0) {
 	    /* do nothing */
@@ -733,9 +811,14 @@ int main(int argc, char **argv)
 		       GT_("couldn't time-check %s (error %d)\n"),
 		       rcfile, errno);
 	}
-	else if (rcstat.st_mtime > parsetime)
+	if (-1 == stat(netrc_file, &netrcstat) && ENOENT != errno) {
+		report(stderr, 
+			GT_("couldn't time-check %s (error %d)\n"),
+			netrc_file, errno);
+	}
+	else if (rcstat.st_mtime > parsetime || netrcstat.st_mtime > netrc_parsetime)
 	{
-	    report(stdout, GT_("restarting fetchmail (%s changed)\n"), rcfile);
+	    report(stdout, GT_("restarting fetchmail (%s changed)\n"), rcstat.st_mtime > parsetime ? rcfile : netrc_file);
 
 	    /* restore the startup directory */
 	    if (!currentwd[0] || chdir (currentwd) == -1)
@@ -764,6 +847,7 @@ int main(int argc, char **argv)
 	     * path wasn't used in the initial call.  (If I recall
 	     * correctly, Linux saves it but many other Unices don't.)
 	     */
+            xfree(netrc_file);
 	    execvp(argv[0], argv);
 	    report(stderr, GT_("attempt to re-exec fetchmail failed\n"));
 	}
@@ -842,40 +926,7 @@ int main(int argc, char **argv)
 			successes++;
 		    else if (!check_only && 
 			     ((querystatus!=PS_NOMAIL) || (outlevel==O_DEBUG)))
-			switch(querystatus)
-			{
-			case PS_SUCCESS:
-			    report(stdout,GT_("Query status=0 (SUCCESS)\n"));break;
-			case PS_NOMAIL: 
-			    report(stdout,GT_("Query status=1 (NOMAIL)\n")); break;
-			case PS_SOCKET:
-			    report(stdout,GT_("Query status=2 (SOCKET)\n")); break;
-			case PS_AUTHFAIL:
-			    report(stdout,GT_("Query status=3 (AUTHFAIL)\n"));break;
-			case PS_PROTOCOL:
-			    report(stdout,GT_("Query status=4 (PROTOCOL)\n"));break;
-			case PS_SYNTAX:
-			    report(stdout,GT_("Query status=5 (SYNTAX)\n")); break;
-			case PS_IOERR:
-			    report(stdout,GT_("Query status=6 (IOERR)\n"));  break;
-			case PS_ERROR:
-			    report(stdout,GT_("Query status=7 (ERROR)\n"));  break;
-			case PS_EXCLUDE:
-			    report(stdout,GT_("Query status=8 (EXCLUDE)\n")); break;
-			case PS_LOCKBUSY:
-			    report(stdout,GT_("Query status=9 (LOCKBUSY)\n"));break;
-			case PS_SMTP:
-			    report(stdout,GT_("Query status=10 (SMTP)\n")); break;
-			case PS_DNS:
-			    report(stdout,GT_("Query status=11 (DNS)\n")); break;
-			case PS_BSMTP:
-			    report(stdout,GT_("Query status=12 (BSMTP)\n")); break;
-			case PS_MAXFETCH:
-			    report(stdout,GT_("Query status=13 (MAXFETCH)\n"));break;
-			default:
-			    report(stdout,GT_("Query status=%d\n"),querystatus);
-			    break;
-			}
+			reportquerystatus(stdout);
 
 #ifdef CAN_MONITOR
 		    if (ctl->server.monitor)
@@ -942,15 +993,16 @@ int main(int argc, char **argv)
 	     * unwedge all servers in case the problem has been
 	     * manually repaired.
 	     */
-	    if ((lastsig = interruptible_idle(run.poll_interval)))
+            lastsig = interruptible_idle(run.poll_interval);
+	    if (lastsig)
 	    {
 		if (outlevel > O_SILENT)
-#ifdef SYS_SIGLIST_DECLARED
-		    report(stdout, 
-		       GT_("awakened by %s\n"), sys_siglist[lastsig]);
+#ifdef HAVE_STRSIGNAL
+		    report(stdout, GT_("awakened by %s\n"), strsignal(lastsig));
+#elif SYS_SIGLIST_DECLARED
+		    report(stdout, GT_("awakened by %s\n"), sys_siglist[lastsig]);
 #else
-	    	    report(stdout, 
-		       GT_("awakened by signal %d\n"), lastsig);
+		    report(stdout, GT_("awakened by signal %d\n"), lastsig);
 #endif
 		for (ctl = querylist; ctl; ctl = ctl->next)
 		    ctl->wedged = FALSE;
@@ -968,6 +1020,8 @@ int main(int argc, char **argv)
 
     terminate_run(0);
 
+    xfree(netrc_file);
+
     if (successes)
 	exit(PS_SUCCESS);
     else if (querystatus)
@@ -1004,12 +1058,14 @@ static void optmerge(struct query *h2, struct query *h1, int force)
 
 #define   FLAG_MERGE(fld) do { if (force ? !!h1->fld : !h2->fld) h2->fld = h1->fld; } while (0)
 #define STRING_MERGE(fld) do { if (force ? !!h1->fld : !h2->fld) { if (h2->fld) free((void *)h2->fld), h2->fld = 0; if (h1->fld) { if (h1->fld != STRING_DISABLED) h2->fld = xstrdup(h1->fld); else h2->fld = STRING_DISABLED; } } } while (0)
+#define STRING_MERGE_SAFE(fld) do { if (force ? !!h1->fld : !h2->fld) { if (h2->fld) { fm_safe_clearcstr((void *)h2->fld); free((void *)h2->fld); h2->fld = 0; } if (h1->fld) { if (h1->fld != STRING_DISABLED) h2->fld = xstrdup(h1->fld); else h2->fld = STRING_DISABLED; } } } while (0)
     STRING_MERGE(server.via);
     FLAG_MERGE(server.protocol);
     STRING_MERGE(server.service);
     FLAG_MERGE(server.interval);
     FLAG_MERGE(server.authenticate);
     FLAG_MERGE(server.timeout);
+    FLAG_MERGE(server.idle_timeout);
     STRING_MERGE(server.envelope);
     FLAG_MERGE(server.envskip);
     STRING_MERGE(server.qvirtual);
@@ -1019,7 +1075,7 @@ static void optmerge(struct query *h2, struct query *h1, int force)
     FLAG_MERGE(server.uidl);
     STRING_MERGE(server.principal);
     STRING_MERGE(server.esmtp_name);
-    STRING_MERGE(server.esmtp_password);
+    STRING_MERGE_SAFE(server.esmtp_password);
 
 #ifdef CAN_MONITOR
     STRING_MERGE(server.interface);
@@ -1034,7 +1090,7 @@ static void optmerge(struct query *h2, struct query *h1, int force)
 
     FLAG_MERGE(wildcard);
     STRING_MERGE(remotename);
-    STRING_MERGE(password);
+    STRING_MERGE_SAFE(password);
     STRING_MERGE(mda);
     STRING_MERGE(bsmtp);
     FLAG_MERGE(listener);
@@ -1044,6 +1100,7 @@ static void optmerge(struct query *h2, struct query *h1, int force)
     STRING_MERGE(postconnect);
 
     FLAG_MERGE(keep);
+    STRING_MERGE(moveto);
     FLAG_MERGE(flush);
     FLAG_MERGE(limitflush);
     FLAG_MERGE(fetchall);
@@ -1055,6 +1112,7 @@ static void optmerge(struct query *h2, struct query *h1, int force)
     FLAG_MERGE(dropdelivered);
     FLAG_MERGE(mimedecode);
     FLAG_MERGE(idle);
+    FLAG_MERGE(forceidle);
     FLAG_MERGE(limit);
     FLAG_MERGE(warnings);
     FLAG_MERGE(fetchlimit);
@@ -1101,6 +1159,7 @@ static int load_params(int argc, char **argv, int optind)
     def_opts.smtpname = (char *)0;
     def_opts.server.protocol = P_AUTO;
     def_opts.server.timeout = CLIENT_TIMEOUT;
+    def_opts.server.idle_timeout = CLIENT_IDLE_TIMEOUT;
     def_opts.server.esmtp_name = user;
     def_opts.server.badheader = BHREJECT;
     def_opts.warnings = WARNING_INTERVAL;
@@ -1131,14 +1190,15 @@ static int load_params(int argc, char **argv, int optind)
     }
 
     /* this builds the host list */
-    if ((st = prc_parse_file(rcfile, !versioninfo)) != 0)
+    if ((st = prc_parse_file(rcfile, !versioninfo)) != 0) {
 	/*
 	 * FIXME: someday, send notification mail here if backgrounded.
 	 * Right now, that can happen if the user changes the rcfile
 	 * while the fetchmail is running in background.  Do similarly
 	 * for the other exit() calls in this function.
 	 */
-	exit(st);
+	// exit(st); // -- deferred, see below
+    }
 
     if ((implicitmode = (optind >= argc)))
     {
@@ -1231,8 +1291,8 @@ static int load_params(int argc, char **argv, int optind)
     if (cmd_run.use_syslog)
 	run.use_syslog = (cmd_run.use_syslog == FLAG_TRUE);
     if (cmd_run.postmaster) {
-    	free((void *)run.postmaster);
-	run.postmaster = cmd_run.postmaster;
+	xfree(run.postmaster);
+	run.postmaster = xstrdup(cmd_run.postmaster);
     }
     if (cmd_run.bouncemail)
 	run.bouncemail = (cmd_run.bouncemail == FLAG_TRUE);
@@ -1318,6 +1378,7 @@ static int load_params(int argc, char **argv, int optind)
 	    DEFAULT(ctl->dropdelivered, FALSE);
 	    DEFAULT(ctl->mimedecode, FALSE);
 	    DEFAULT(ctl->idle, FALSE);
+	    DEFAULT(ctl->forceidle, FALSE);
 	    DEFAULT(ctl->server.dns, TRUE);
 	    DEFAULT(ctl->server.uidl, FALSE);
 	    DEFAULT(ctl->use_ssl, FALSE);
@@ -1369,7 +1430,7 @@ static int load_params(int argc, char **argv, int optind)
 	    /* if `user' doesn't name a real local user, try to run as root */
 	    if ((pw = getpwnam(user)) == (struct passwd *)NULL)
 		ctl->uid = 0;
-            else
+	    else
 		ctl->uid = pw->pw_uid;	/* for local delivery via MDA */
 	    if (!ctl->localnames)	/* for local delivery via SMTP */
 		save_str_pair(&ctl->localnames, user, NULL);
@@ -1419,15 +1480,17 @@ static int load_params(int argc, char **argv, int optind)
 				   ctl->server.pollname);
 		    exit(PS_SYNTAX);
 		}
-		switch (ctl->server.protocol) {
+		if (!ctl->server.plugin) { // do not check this under plugin, we don't know what protocol the plugin talks, Debian Bug#1076604
+		    switch (ctl->server.protocol) {
 			case P_POP3: case P_APOP:
-				if (port == 995 && !ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need --ssl for port 995/service pop3s.\n"), ctl->server.pollname);
-				if (port == 110 &&  ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need port 995/service pop3s for --ssl.\n"), ctl->server.pollname);
-				break;
+			    if (port == 995 && !ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need --ssl for port 995/service pop3s.\n"), ctl->server.pollname);
+			    if (port == 110 &&  ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need port 995/service pop3s for --ssl.\n"), ctl->server.pollname);
+			    break;
 			case P_IMAP:
-				if (port == 993 && !ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need --ssl for port 993/service imaps.\n"), ctl->server.pollname);
-				if (port == 143 &&  ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need port 993/service imaps for --ssl.\n"), ctl->server.pollname);
-				break;
+			    if (port == 993 && !ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need --ssl for port 993/service imaps.\n"), ctl->server.pollname);
+			    if (port == 143 &&  ctl->use_ssl) report(stderr, GT_("WARNING: %s configuration invalid, you normally need port 993/service imaps for --ssl.\n"), ctl->server.pollname);
+			    break;
+		    }
 		}
 	    }
 	    if (ctl->listener == LMTP_MODE)
@@ -1479,15 +1542,24 @@ static int load_params(int argc, char **argv, int optind)
     if (!run.postmaster)
     {
 	if (getuid() != ROOT_UID)		/* ordinary user */
-	    run.postmaster = user;
+	    run.postmaster = xstrdup(user);
 	else					/* root */
-	    run.postmaster = "postmaster";
+	    run.postmaster = xstrdup("postmaster");
+    }
+
+    if (st) {
+	if (outlevel >= O_DEBUG) {
+	    report(stderr, GT_("cannot parse rcfile, debug mode: dumping obtained configuration as Python:\n"));
+	    puts("raise Exception(\"This output is not meant to be processed by fetchmailrc.\")");
+	    dump_config(&run, querylist);
+	}
+	exit(PS_SYNTAX);
     }
 
     return(implicitmode);
 }
 
-static RETSIGTYPE terminate_poll(int sig)
+static void terminate_poll(int sig)
 /* to be executed at the end of a poll cycle */
 {
 
@@ -1505,11 +1577,14 @@ static RETSIGTYPE terminate_poll(int sig)
 #endif /* POP3_ENABLE */
 }
 
-static RETSIGTYPE terminate_run(int sig)
-/* to be executed on normal or signal-induced termination */
+static void terminate_run(int sig)
+/* to be executed on normal or signal-induced termination
+ * - will not return if sig != 0, but log the signal and exit. */
 {
     struct query	*ctl;
 
+    clear_free_visbuf();
+
     terminate_poll(sig);
 
     /* 
@@ -1525,16 +1600,17 @@ static RETSIGTYPE terminate_run(int sig)
      */
     for (ctl = querylist; ctl; ctl = ctl->next)
 	if (ctl->password)
-	  memset(ctl->password, '\0', strlen(ctl->password));
+	  // revised on 2025-09-09 for fetchmail 6.5.5 release
+	  // compilers are free to optimize away a memset(),
+	  // se we need to use something else.
+	  fm_safe_clearcstr(ctl->password);
 
-#if !defined(HAVE_ATEXIT)
-    fm_lock_release();
-#endif
-
-    if (activecount == 0)
-	exit(PS_NOMAIL);
-    else
-	exit(successes ? PS_SUCCESS : querystatus);
+    if (sig) {
+	if (activecount == 0)
+	    exit(PS_NOMAIL);
+	else
+	    exit(successes ? PS_SUCCESS : querystatus);
+    }
 }
 
 /*
@@ -1674,10 +1750,8 @@ static void dump_params (struct runctl *runp,
 	printf(GT_("Logfile is %s\n"), runp->logfile);
     if (strcmp(runp->idfile, IDFILE_NAME))
 	printf(GT_("Idfile is %s\n"), runp->idfile);
-#if defined(HAVE_SYSLOG)
     if (runp->use_syslog)
 	printf(GT_("Progress messages will be logged via syslog\n"));
-#endif
     if (runp->invisible)
 	printf(GT_("Fetchmail will masquerade and will not generate Received\n"));
     if (runp->showdots)
@@ -1780,7 +1854,7 @@ static void dump_params (struct runctl *runp,
 	case A_KERBEROS_V5:
 	    printf(GT_("  Kerberos V5 authentication will be forced.\n"));
 	    break;
-	case A_SSH:
+	case A_IMPLICIT:
 	    printf(GT_("  End-to-end encryption assumed.\n"));
 	    break;
 	}
@@ -1810,9 +1884,14 @@ static void dump_params (struct runctl *runp,
 	if (ctl->server.timeout > 0)
 	    printf(GT_("  Server nonresponse timeout is %d seconds"), ctl->server.timeout);
 	if (ctl->server.timeout ==  CLIENT_TIMEOUT)
-	    printf(GT_(" (default).\n"));
-	else
-	    printf(".\n");
+	    printf(GT_(" (default)"));
+	printf(".\n");
+
+	if (ctl->server.idle_timeout > 0)
+	    printf(GT_("  Server IDLE timeout is %d seconds"), ctl->server.idle_timeout);
+	if (ctl->server.idle_timeout ==  CLIENT_IDLE_TIMEOUT)
+	    printf(GT_(" (default)"));
+	 printf(".\n");
 
 	if (MAILBOX_PROTOCOL(ctl)) 
 	{
@@ -1833,6 +1912,10 @@ static void dump_params (struct runctl *runp,
 	    printf(ctl->keep
 		   ? GT_("  Fetched messages will be kept on the server (--keep on).\n")
 		   : GT_("  Fetched messages will not be kept on the server (--keep off).\n"));
+	    if (ctl->moveto != NULL)
+		printf(GT_("  Message will be moved to %s folder and won't be deleted.\n"), ctl->moveto);
+	    else
+		printf(GT_("  Message will be deleted (no --moveto).\n"));
 	    printf(ctl->flush
 		   ? GT_("  Old messages will be flushed before message retrieval (--flush on).\n")
 		   : GT_("  Old messages will not be flushed before message retrieval (--flush off).\n"));
@@ -1857,6 +1940,9 @@ static void dump_params (struct runctl *runp,
 	    printf(ctl->idle
 		   ? GT_("  Idle after poll is enabled (idle on).\n")
 		   : GT_("  Idle after poll is disabled (idle off).\n"));
+	    printf(ctl->forceidle
+		   ? GT_("  Idle after poll is forced (forceidle on).\n")
+		   : GT_("  Idle after poll is not forced (forceidle off).\n"));
 	    printf(ctl->dropstatus
 		   ? GT_("  Nonempty Status lines will be discarded (dropstatus on)\n")
 		   : GT_("  Nonempty Status lines will be kept (dropstatus off)\n"));
@@ -2006,7 +2092,7 @@ static void dump_params (struct runctl *runp,
 			       : GT_("  DNS lookup for multidrop addresses is disabled.\n"));
 			if (ctl->server.dns)
 			{
-	       		    if (ctl->server.checkalias)
+			    if (ctl->server.checkalias)
 				printf(GT_("  Server aliases will be compared with multidrop addresses by IP address.\n"));
 			    else
 				printf(GT_("  Server aliases will be compared with multidrop addresses by name.\n"));
