@@ -30,6 +30,7 @@
 #include <sys/time.h>
 #include <time.h>
 #include <limits.h>
+#include <assert.h>
 
 #include "socket.h"
 #include "i18n.h"
@@ -66,12 +67,14 @@ static char **parse_plugin(const char *plugin, const char *host, const char *ser
 	char **argvec;
 	const char *c, *p;
 	char *cp, *plugin_copy;
-	unsigned int plugin_copy_len;
-	unsigned int plugin_offset = 0, plugin_copy_offset = 0;
-	unsigned int i, vecsiz = 2 * sizeof(char*), host_count = 0, service_count = 0;
-	unsigned int plugin_len = strlen(plugin);
-	unsigned int host_len = strlen(host);
-	unsigned int service_len = strlen(service);
+	size_t plugin_copy_len;
+	size_t plugin_offset = 0, plugin_copy_offset = 0;
+	unsigned int i;
+	size_t vecsiz = 2 * sizeof(char*);
+	unsigned int host_count = 0, service_count = 0;
+	size_t plugin_len = strlen(plugin);
+	size_t host_len = strlen(host);
+	size_t service_len = strlen(service);
 
 	for (c = p = plugin; *c; c++)
 	{	if (isspace((unsigned char)*c) && !isspace((unsigned char)*p))
@@ -325,16 +328,102 @@ int SockOpen(const char *host, const char *service,
     return i;
 }
 
+static size_t fmtbufsize;
+static char *fmtbuf;
 
-int SockPrintf(int sock, const char* format, ...)
+void SockFreeFbuf(void) {
+    if (fmtbuf) {
+	xzerofree(fmtbuf);
+	fmtbufsize = 0;
+    }
+}
+
+/** round_pow2 rounds its argument up to the next power of 2.
+ * Returns UINT_MAX on overflow. */
+static unsigned int round_pow2(unsigned int v, bool *errorflag) {
+    if (!v) return 1;  // 2^0 = 1
+#ifndef NDEBUG
+    unsigned int orig_v = v;
+#endif
+    v--;
+    v |= v >> 1;
+    v |= v >> 2;
+    v |= v >> 4;
+    v |= v >> 8;
+    v |= v >> 16;
+#if SIZEOF_INT > 4
+    v |= v >> 32;
+#endif
+#if SIZEOF_INT > 8 // OK - that will get interesting
+    v |= v >> 64;
+#endif
+    v++;
+    if (!v) {
+	v = UINT_MAX;
+	if (errorflag) *errorflag = true;
+    }
+#ifndef NDEBUG
+    assert(v >= orig_v);
+#endif
+    return v;
+}
+
+/** allocate a sufficiently sized fmtbuf, keep its size in fmtbufsize,
+ * format the arguments as if by fprintf, and use SockWrite to ship them
+ * to socket \sock.
+ * Optionally clear the internal formatting buffer to not let secrets
+ * around.
+ * Will abort() if vsnprintf() malfunctions. */
+__attribute__ ((format(printf, 3, 0)))
+static ssize_t SockVPrintf(int sock, bool clearbuf /** if true, clear internal buffer after sending the data */, const char *format, va_list args)
+{
+    va_list args_save;
+    va_copy(args_save, args);
+    int somesize = vsnprintf(NULL, 0, format, args_save);
+    va_end(args_save);
+    if (somesize < 0) return somesize; /* ERROR */
+    // we're now sure that somesize isn't negative
+    ++somesize; // make room for NUL byte
+
+    enum { MINFMTSIZE = 4096 };
+    if (fmtbufsize < (unsigned)somesize) {
+	// if (somesize < MINFMTSIZE) somesize = MINFMTSIZE;
+	fmtbufsize = round_pow2(somesize, NULL);  // this shouldn't usually trigger
+	fmtbuf = (char *)xrealloc(fmtbuf, fmtbufsize);
+    }
+    somesize = vsnprintf(fmtbuf, fmtbufsize, format, args);
+    if (somesize < 0 || (unsigned)somesize >= fmtbufsize) {
+	// paranoia - this indicates an error or an overflow
+	// neither should happen because we should
+	// have allocated sufficient memory and it's
+	// implausible for vsnprintf to fail this time
+	// if it succeeded the first time.
+	fm_safe_clearmem(fmtbuf, fmtbufsize);
+	return -1;
+    }
+    ssize_t retval = SockWrite(sock, fmtbuf, somesize);
+    if (clearbuf) fm_safe_clearmem(fmtbuf, fmtbufsize);
+    return retval;
+}
+
+ssize_t SockPrintfClear(int sock, const char* format, ...)
 {
     va_list ap;
-    char buf[8192];
 
-    va_start(ap, format) ;
-    vsnprintf(buf, sizeof(buf), format, ap);
+    va_start(ap, format);
+    ssize_t result = SockVPrintf(sock, true, format, ap);
     va_end(ap);
-    return SockWrite(sock, buf, strlen(buf));
+    return result;
+}
+
+ssize_t SockPrintf(int sock, const char* format, ...)
+{
+    va_list ap;
+
+    va_start(ap, format);
+    ssize_t result = SockVPrintf(sock, false, format, ap);
+    va_end(ap);
+    return result;
 }
 
 #ifdef SSL_ENABLE
@@ -354,7 +443,6 @@ enum { SSL_min_security_level = 2 };
 #define DO_SSL_VERSION_CHECK 1
 #ifdef LIBRESSL_VERSION_NUMBER
 # ifdef __OpenBSD__
-#  pragma message "WARNING - Linking against LibreSSL, which is not a supported configuration."
 #  define TLSprovider "LibreSSL(OpenBSD)"
 #  undef  DO_SSL_VERSION_CHECK
 #  define DO_SSL_VERSION_CHECK 0
@@ -433,12 +521,11 @@ static SSL *SSLGetContext(int);
 static int check_sslfp(const X509 *x509_cert, char *text, size_t textsize, int *retFlag);
 #endif /* SSL_ENABLE */
 
-int SockWrite(int sock, const char *buf, int len)
+ssize_t SockWrite(int sock, const char *buf, size_t len)
 {
-	/* WARNING: This is messy - we use an int pointer, so does SSL_write, but fm_write/write would use ssize_t instead.
-	Assume that write() won't ever return something that wouldn't fit an int. */
-    int n;
-    int wrlen = 0;
+    ssize_t n;
+    ssize_t wrlen = 0;
+    if (len > INT_MAX) len = INT_MAX;  // clamp to sane default
 #ifdef	SSL_ENABLE
     SSL *ssl = SSLGetContext(sock);
 #endif
@@ -448,13 +535,13 @@ int SockWrite(int sock, const char *buf, int len)
 #ifdef SSL_ENABLE
 	if (ssl) {
 	    ERR_clear_error();
-	    n = SSL_write(ssl, buf, len);
+	    n = SSL_write(ssl, buf, (int)len);
 	} else
 #endif /* SSL_ENABLE */
-	    n = fm_write(sock, buf, len);
+	    n = (int)fm_write(sock, buf, len); // since len is an int, we can cast the return value to int
 #ifdef SSL_ENABLE
-        if (n <= 0 && ssl != NULL) {
-	    int ssle = SSL_get_error(ssl, n); // do this before flushing the error queue!
+        if (n <= 0 && ssl) {
+	    int ssle = SSL_get_error(ssl, (int)n); // do this before flushing the error queue!
 	    // map error code to n = 0 -> retryable or n = -1 -> true error
 	    switch(ssle) {
 		case SSL_ERROR_ZERO_RETURN:
@@ -478,10 +565,14 @@ int SockWrite(int sock, const char *buf, int len)
     return wrlen;
 }
 
-int SockRead(int sock, char *buf, int len)
+ssize_t SockRead(int sock, char *buf, size_t len)
 {
+    if (len > INT_MAX) len = INT_MAX;  // clamp to sane default
+    // as for SockWrite, we cast some ssize_t return values to int
+    // on the assumption that we can't possibly read more than INT_MAX
+    // because our input type already is only an int.
     char *newline, *bp = buf;
-    int n;
+    ssize_t n;
 #ifdef	SSL_ENABLE
     SSL *ssl;
 #endif
@@ -516,11 +607,11 @@ int SockRead(int sock, char *buf, int len)
 			later change the behavior of SSL_peek
 			to "fix" this problem...  :-(	*/
 		ERR_clear_error();
-		if ((n = SSL_peek(ssl, bp, len)) <= 0) {
+		if ((n = SSL_peek(ssl, bp, (int)len)) <= 0) {
 			/* SSL_peek says no data...  Does he mean no data
 			or did the connection blow up?  If we got an error
 			then bail! */
-			e = SSL_get_error(ssl, n);
+			e = SSL_get_error(ssl, (int)n);
 			if (SSL_ERROR_NONE != e
 #ifdef USING_WOLFSSL
 			/* wolfSSL 5.0.0 may return SSL_ERROR_WANT_READ when
@@ -541,14 +632,17 @@ int SockRead(int sock, char *buf, int len)
 			 * the strchr at this point yet */
 			newline = NULL;
 		} else if ((newline = (char *)memchr(bp, '\n', n)) != NULL)
-			n = newline - bp + 1;
+			n = (int)(newline - bp + 1);
 		/* Matthias Andree: SSL_read can return 0, in that case
 		 * we must call SSL_get_error to figure if there was
 		 * an error or just a "no data" condition */
-		// OpenSSL 3 documents "Old documentation indicated a difference between 0 and -1, and that -1 was retryable. You should instead call SSL_get_error() to find out if it's retryable."
+		// OpenSSL 3 documents "Old documentation indicated a difference between 0 and -1, and that -1 was retryable.
+                // You should instead call SSL_get_error() to find out if it's retryable."
+                // However, we've set auto retry mode on the SSL_CTX so we needn't care about the
+                // SSL_ERROR_WANT_READ SSL_ERROR_WANT_WRITE return codes.
 		ERR_clear_error();
-		if ((n = SSL_read(ssl, bp, n)) <= 0) {
-			e = SSL_get_error(ssl, n);
+		if ((n = SSL_read(ssl, bp, (int)n)) <= 0) {
+			e = SSL_get_error(ssl, (int)n);
 			if (SSL_ERROR_NONE != e) {
 				ERR_print_errors_fp(stderr);
 				return -1;
@@ -568,11 +662,11 @@ int SockRead(int sock, char *buf, int len)
 	{
 	    if (sock < 0)
 		return -1;
-	    if ((n = fm_peek(sock, bp, len)) <= 0)
+	    if ((n = (int)fm_peek(sock, bp, len)) <= 0)
 		return -1;
 	    if ((newline = (char *)memchr(bp, '\n', n)) != NULL)
-		n = newline - bp + 1;
-	    if ((n = fm_read(sock, bp, n)) == -1)
+		n = (int)(newline - bp + 1);
+	    if ((n = (int)fm_read(sock, bp, n)) == -1)
 		return -1;
 	}
 	bp += n;
@@ -580,7 +674,7 @@ int SockRead(int sock, char *buf, int len)
     } while (!newline && len);
     *bp = '\0';
 
-    return bp - buf;
+    return (int)(bp - buf);
 }
 
 int SockPeek(int sock)
@@ -617,7 +711,8 @@ int SockPeek(int sock)
 	}
 	else
 #endif /* SSL_ENABLE */
-	    n = sock >= 0 ? fm_peek(sock, &ch, 1) : -1;
+	    /* casting to int is safe on the next line, 1 fits an int */
+	       n = sock >= 0 ? (int)fm_peek(sock, &ch, 1) : -1;
 	if (n == -1)
 		return -1;
 
@@ -628,7 +723,7 @@ int SockPeek(int sock)
 
 static	char *_ssl_server_cname = NULL;
 static	int _check_fp;
-static	char *_check_digest;
+static	const char *_check_digest;
 static	char *_server_label;
 static	int _depth0ck;
 static	int _firstrun;
@@ -863,7 +958,7 @@ int check_sslfp(const X509 *x509_cert, char *text, size_t textsize, int *retFlag
         fm_EVP_MD_const EVP_MD *digest_impl = NULL;
 	unsigned char digest[EVP_MAX_MD_SIZE];
 	char *dig = NULL;
-	char *expected_digest = _check_digest;
+	const char *expected_digest = _check_digest;
 
 	if (_check_digest && 1 == match_regex("^\\{[a-z0-9-]+\\}", _check_digest)) {
 		size_t len = strcspn(_check_digest + 1, "}");
@@ -904,6 +999,7 @@ int check_sslfp(const X509 *x509_cert, char *text, size_t textsize, int *retFlag
 	if (!success)
         {
                 report(stderr, GT_("X509_digest(): Out of memory!\n"));
+		xfree(dig);
                 return 0;
         }
 
@@ -971,6 +1067,10 @@ static const char *SSLCertGetCN(const char *mycert,
 	X509       *x509_cert = NULL;
 	X509_NAME  *certname  = NULL;
 
+	// limit namebufferlen to what OpenSSL can handle
+	// in X509_NAME_get_text_by_NID
+	if (namebufferlen > INT_MAX) namebufferlen = INT_MAX;
+
 	if (namebuffer && namebufferlen > 0) {
 		namebuffer[0] = 0x00;
 		certBio = BIO_new_file(mycert,"r");
@@ -982,7 +1082,7 @@ static const char *SSLCertGetCN(const char *mycert,
 			certname = X509_get_subject_name(x509_cert);
 			if (certname &&
 			    X509_NAME_get_text_by_NID(certname, NID_commonName,
-						      namebuffer, namebufferlen) > 0)
+						      namebuffer, (int)namebufferlen) > 0)
 				ret = namebuffer;
 			X509_free(x509_cert);
 		}
@@ -1053,7 +1153,7 @@ static int OSSL_proto_version_logic(int sock, const char **myproto)
 /* flush (discard) pending input from socket
  * if this doesn't work with FIONREAD, we can instead rewrite to loop over 
  * select() and then read()/recv(). */
-void inputflush(int sock) {
+static void inputflush(int sock) {
 	char buf[1024];
 	int s;
 
@@ -1098,7 +1198,7 @@ static char *fm_ssl_all_errors(void) {
  */
 int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck,
     char *cacertfile, char *certpath,
-    char *fingerprint, char *servercname, char *label, char **remotename)
+    const char *fingerprint, char *servercname, char *label, char **remotename)
 {
         struct stat randstat;
         int i;
@@ -1238,7 +1338,7 @@ int SSLOpen(int sock, char *mycert, char *mykey, const char *myproto, int certck
 			} else if (outlevel >= O_DEBUG) {
 				report(stdout, GT_("Parsed %s to set new security level %d\n"), nseclv, (int)seclvl);
 			}
-			SSL_CTX_set_security_level(_ctx[sock], seclvl);
+			SSL_CTX_set_security_level(_ctx[sock], (int)seclvl); // seclvl must be range-checked for INT_MAX above
 		} else {
 			if (SSL_CTX_get_security_level(_ctx[sock]) < SSL_min_security_level) {
 				SSL_CTX_set_security_level(_ctx[sock], SSL_min_security_level); /* void function */
